@@ -1,4 +1,5 @@
 """Subprocess specification and related utilities."""
+
 import contextlib
 import inspect
 import io
@@ -11,13 +12,13 @@ import stat
 import subprocess
 import sys
 
-import xonsh.environ as xenv
-import xonsh.jobs as xj
-import xonsh.lazyasd as xl
-import xonsh.lazyimps as xli
+import xonsh.lib.lazyasd as xl
+import xonsh.lib.lazyimps as xli
 import xonsh.platform as xp
+import xonsh.procs.jobs as xj
 import xonsh.tools as xt
 from xonsh.built_ins import XSH
+from xonsh.procs.executables import locate_executable
 from xonsh.procs.pipelines import (
     STDOUT_CAPTURE_KINDS,
     CommandPipeline,
@@ -133,7 +134,7 @@ def get_script_subproc_command(fname, args):
 @xl.lazyobject
 def _REDIR_REGEX():
     name = r"(o(?:ut)?|e(?:rr)?|a(?:ll)?|&?\d?)"
-    return re.compile("{r}(>?>|<){r}$".format(r=name))
+    return re.compile(f"{name}(>?>|<){name}$")
 
 
 @xl.lazyobject
@@ -171,10 +172,6 @@ def _O2E_MAP():
     return frozenset({f"{o}>{e}" for e in _REDIR_ERR for o in _REDIR_OUT if o != ""})
 
 
-def _is_redirect(x):
-    return isinstance(x, str) and _REDIR_REGEX.match(x)
-
-
 def safe_open(fname, mode, buffering=-1):
     """Safely attempts to open a file in for xonsh subprocs."""
     # file descriptors
@@ -226,6 +223,9 @@ def _parse_redirects(r, loc=None):
 
 def _redirect_streams(r, loc=None):
     """Returns stdin, stdout, stderr tuple of redirections."""
+    if isinstance(loc, list):
+        raise Exception(f"Unsupported redirect: {r!r} {loc!r}")
+
     stdin = stdout = stderr = None
     no_ampersand = r.replace("&", "")
     # special case of redirecting stderr to stdout
@@ -253,6 +253,17 @@ def _redirect_streams(r, loc=None):
     return stdin, stdout, stderr
 
 
+def _flatten_cmd_redirects(cmd):
+    """Transforms a command like ['ls', ('>', '/dev/null')] into ['ls', '>', '/dev/null']."""
+    new_cmd = []
+    for c in cmd:
+        if isinstance(c, tuple):
+            new_cmd.extend(c)
+        else:
+            new_cmd.append(c)
+    return new_cmd
+
+
 def default_signal_pauser(n, f):
     """Pauses a signal, as needed."""
     signal.pause()
@@ -266,6 +277,45 @@ def no_pg_xonsh_preexec_fn():
     signal.signal(signal.SIGTSTP, default_signal_pauser)
 
 
+class DecoratorAlias:
+    """Decorator alias base class."""
+
+    descr = "DecoratorAlias base."
+
+    def __call__(
+        self,
+        args,
+        stdin=None,
+        stdout=None,
+        stderr=None,
+        spec=None,
+        stack=None,
+        **kwargs,
+    ):
+        print(self.descr, file=stdout)
+
+    def decorate_spec(self, spec):
+        """Modify spec immediately after modifier added."""
+        pass
+
+    def decorate_spec_pre_run(self, pipeline, spec, spec_num):
+        """Modify spec before run."""
+        pass
+
+
+class SpecAttrDecoratorAlias(DecoratorAlias):
+    """Decorator Alias for spec attributes."""
+
+    def __init__(self, set_attributes: dict, descr=""):
+        self.set_attributes = set_attributes
+        self.descr = descr
+        super().__init__()
+
+    def decorate_spec(self, spec):
+        for a, v in self.set_attributes.items():
+            setattr(spec, a, v)
+
+
 class SubprocSpec:
     """A container for specifying how a subprocess command should be
     executed.
@@ -355,7 +405,7 @@ class SubprocSpec:
         else:
             self.env = None
         # pure attrs
-        self.args = list(cmd)
+        self.args = _flatten_cmd_redirects(cmd)
         self.alias = None
         self.alias_name = None
         self.alias_stack = XSH.env.get("__ALIAS_STACK", "").split(":")
@@ -363,11 +413,15 @@ class SubprocSpec:
         self.is_proxy = False
         self.background = False
         self.threadable = True
+        self.force_threadable = None  # Set this value to ignore threadable prediction.
         self.pipeline_index = None
         self.last_in_pipeline = False
         self.captured_stdout = None
         self.captured_stderr = None
         self.stack = None
+        self.decorators = []  # List of DecoratorAlias objects that applied to spec.
+        self.output_format = XSH.env.get("XONSH_SUBPROC_OUTPUT_FORMAT", "stream_lines")
+        self.raise_subproc_error = None  # Spec-based $RAISE_SUBPROC_ERROR.
 
     def __str__(self):
         s = self.__class__.__name__ + "(" + str(self.cmd) + ", "
@@ -400,7 +454,7 @@ class SubprocSpec:
         else:
             safe_close(value)
             msg = "Multiple inputs for stdin for {0!r}"
-            msg = msg.format(" ".join(self.args))
+            msg = msg.format(self.get_command_str())
             raise xt.XonshError(msg)
 
     @property
@@ -416,7 +470,7 @@ class SubprocSpec:
         else:
             safe_close(value)
             msg = "Multiple redirections for stdout for {0!r}"
-            msg = msg.format(" ".join(self.args))
+            msg = msg.format(self.get_command_str())
             raise xt.XonshError(msg)
 
     @property
@@ -432,9 +486,12 @@ class SubprocSpec:
         else:
             safe_close(value)
             msg = "Multiple redirections for stderr for {0!r}"
-            msg = msg.format(" ".join(self.args))
+            msg = msg.format(self.get_command_str())
             raise xt.XonshError(msg)
 
+    def get_command_str(self):
+        return " ".join(arg for arg in self.args)
+
     #
     # Execution methods
     #
@@ -477,10 +534,18 @@ class SubprocSpec:
         except FileNotFoundError as ex:
             cmd0 = self.cmd[0]
             if len(self.cmd) == 1 and cmd0.endswith("?"):
-                with contextlib.suppress(OSError):
-                    return self.cls(
-                        ["man", cmd0.rstrip("?")], bufsize=bufsize, **kwargs
+                cmdq = cmd0.rstrip("?")
+                if cmdq in XSH.aliases:
+                    alias = XSH.aliases[cmdq]
+                    descr = (
+                        repr(alias) + (":\n" + doc)
+                        if (doc := getattr(alias, "__doc__", ""))
+                        else ""
                     )
+                    return self.cls(["echo", descr], bufsize=bufsize, **kwargs)
+                else:
+                    with contextlib.suppress(OSError):
+                        return self.cls(["man", cmdq], bufsize=bufsize, **kwargs)
             e = f"xonsh: subprocess mode: command not found: {repr(cmd0)}"
             env = XSH.env
             sug = xt.suggest_commands(cmd0, env)
@@ -548,6 +613,9 @@ class SubprocSpec:
         if events.exists(event_name):
             event = getattr(events, event_name)
             event.fire(spec=self)
+        if events.exists("on_pre_spec_run"):
+            event = events.on_pre_spec_run
+            event.fire(spec=self)
 
     def _post_run_event_fire(self, name, proc):
         events = XSH.builtins.events
@@ -555,6 +623,9 @@ class SubprocSpec:
         if events.exists(event_name):
             event = getattr(events, event_name)
             event.fire(spec=self, proc=proc)
+        if events.exists("on_post_spec_run"):
+            event = events.on_post_spec_run
+            event.fire(spec=self)
 
     #
     # Building methods
@@ -571,10 +642,9 @@ class SubprocSpec:
         # modifications that do not alter cmds may come before creating instance
         spec = kls(cmd, cls=cls, **kwargs)
         # modifications that alter cmds must come after creating instance
-        # perform initial redirects
-        spec.redirect_leading()
-        spec.redirect_trailing()
-        # apply aliases
+        spec.resolve_decorators()  # keep this first
+        spec.resolve_args_list()
+        spec.resolve_redirects()
         spec.resolve_alias()
         spec.resolve_binary_loc()
         spec.resolve_auto_cd()
@@ -583,57 +653,104 @@ class SubprocSpec:
         spec.resolve_stack()
         return spec
 
-    def redirect_leading(self):
-        """Manage leading redirects such as with '< input.txt COMMAND'."""
-        while len(self.cmd) >= 3 and self.cmd[0] == "<":
-            self.stdin = safe_open(self.cmd[1], "r")
-            self.cmd = self.cmd[2:]
+    def add_decorator(self, mod: DecoratorAlias):
+        """Add spec modifier to the specification."""
+        mod.decorate_spec(self)
+        self.decorators.append(mod)
 
-    def redirect_trailing(self):
-        """Manages trailing redirects."""
-        while True:
-            cmd = self.cmd
-            if len(cmd) >= 3 and _is_redirect(cmd[-2]):
-                streams = _redirect_streams(cmd[-2], cmd[-1])
-                self.stdin, self.stdout, self.stderr = streams
-                self.cmd = cmd[:-2]
-            elif len(cmd) >= 2 and _is_redirect(cmd[-1]):
-                streams = _redirect_streams(cmd[-1])
-                self.stdin, self.stdout, self.stderr = streams
-                self.cmd = cmd[:-1]
+    def resolve_decorators(self):
+        """Apply decorators."""
+        if (ln := len(self.cmd)) == 1:
+            return
+        for i in range(ln):
+            c = self.cmd[i]
+            if c in XSH.aliases and isinstance(mod := XSH.aliases[c], DecoratorAlias):
+                self.add_decorator(mod)
             else:
                 break
+        self.cmd = self.cmd[i:]
+
+    def resolve_args_list(self):
+        """Weave a list of arguments into a command."""
+        resolved_cmd = []
+        for c in self.cmd:
+            if (
+                isinstance(c, tuple)
+                and len(c) == 2
+                and isinstance(c[1], list)
+                and len(c[1]) == 1
+            ):
+                # Redirect case e.g. `> file`
+                resolved_cmd.append(
+                    (
+                        c[0],
+                        c[1][0],
+                    )
+                )
+            else:
+                resolved_cmd += c if isinstance(c, list) else [c]
+        self.cmd = resolved_cmd
+
+    def resolve_redirects(self):
+        """Manages redirects."""
+        new_cmd = []
+        for c in self.cmd:
+            if isinstance(c, tuple):
+                streams = _redirect_streams(*c)
+                self.stdin, self.stdout, self.stderr = streams
+            else:
+                new_cmd.append(c)
+        self.cmd = new_cmd
 
     def resolve_alias(self):
-        """Sets alias in command, if applicable."""
+        """Resolving alias and setting up command."""
         cmd0 = self.cmd[0]
-
         if cmd0 in self.alias_stack:
             # Disabling the alias resolving to prevent infinite loop in call stack
-            # and futher using binary_loc to resolve the alias name.
+            # and further using binary_loc to resolve the alias name.
             self.alias = None
             return
 
         if callable(cmd0):
-            alias = cmd0
+            self.alias = cmd0
         else:
-            alias = XSH.aliases.get(cmd0, None)
+            decorators = []
+            if isinstance(XSH.aliases, dict):
+                # Windows tests
+                alias = XSH.aliases.get(cmd0, None)
+                if alias is not None:
+                    alias = alias + self.cmd[1:]
+            else:
+                alias = XSH.aliases.get(
+                    self.cmd,
+                    None,
+                    decorators=decorators,
+                )
             if alias is not None:
                 self.alias_name = cmd0
-        self.alias = alias
+                if callable(alias[0]):
+                    # E.g. `alias == [FuncAlias({'name': 'cd'}), '/tmp']`
+                    self.alias = alias[0]
+                    self.cmd = [cmd0] + alias[1:]
+                else:
+                    # E.g. `alias == ['ls', '-la']`
+                    self.alias = alias
+
+            for mod in decorators:
+                self.add_decorator(mod)
 
     def resolve_binary_loc(self):
         """Sets the binary location"""
         alias = self.alias
         if alias is None:
             cmd0 = self.cmd[0]
-            binary_loc = xenv.locate_binary(cmd0)
-            if binary_loc == cmd0 and cmd0 in self.alias_stack:
+            binary_loc = locate_executable(cmd0)
+            if binary_loc is None and cmd0 and cmd0 in self.alias_stack:
                 raise Exception(f'Recursive calls to "{cmd0}" alias.')
         elif callable(alias):
             binary_loc = None
         else:
-            binary_loc = xenv.locate_binary(alias[0])
+            binary_loc = locate_executable(alias[0])
         self.binary_loc = binary_loc
 
     def resolve_auto_cd(self):
@@ -658,10 +775,8 @@ class SubprocSpec:
             self.cmd.pop(0)
             return
         else:
-            self.cmd = alias + self.cmd[1:]
-            # resolve any redirects the aliases may have applied
-            self.redirect_leading()
-            self.redirect_trailing()
+            self.cmd = alias
+            self.resolve_redirects()
         if self.binary_loc is None:
             return
         try:
@@ -674,27 +789,17 @@ class SubprocSpec:
 
     def resolve_alias_cls(self):
         """Determine which proxy class to run an alias with."""
-        alias = self.alias
-        if not callable(alias):
-            return
-        self.is_proxy = True
-        env = XSH.env
-        thable = env.get("THREAD_SUBPROCS") and getattr(
-            alias, "__xonsh_threadable__", True
-        )
-        cls = ProcProxyThread if thable else ProcProxy
-        self.cls = cls
-        self.threadable = thable
-        # also check capturability, while we are here
-        cpable = getattr(alias, "__xonsh_capturable__", self.captured)
-        self.captured = cpable
+        if callable(self.alias):
+            self.is_proxy = True
+            _update_proc_alias_threadable(self)
+            _update_proc_alias_captured(self)
 
     def resolve_stack(self):
         """Computes the stack for a callable alias's call-site, if needed."""
         if not callable(self.alias):
             return
         # check that we actual need the stack
-        sig = inspect.signature(self.alias)
+        sig = inspect.signature(getattr(self.alias, "func", self.alias))
         if len(sig.parameters) <= 5 and "stack" not in sig.parameters:
             return
         # compute the stack, and filter out these build methods
@@ -702,7 +807,8 @@ class SubprocSpec:
         # we want to filter out one up, e.g. subproc_captured_hiddenobject()
         # after that the stack from the call site starts.
         stack = inspect.stack(context=0)
-        assert stack[3][3] == "run_subproc", "xonsh stack has changed!"
+        if not stack[3][3].startswith("test_"):
+            assert stack[3][3] == "run_subproc", "xonsh stack has changed!"
         del stack[:5]
         self.stack = stack
 
@@ -733,32 +839,55 @@ def _safe_pipe_properties(fd, use_tty=False):
 
 
 def _update_last_spec(last):
-    env = XSH.env
-    captured = last.captured
     last.last_in_pipeline = True
-    if not captured:
+
+    if not last.captured:
         return
-    callable_alias = callable(last.alias)
-    if callable_alias:
-        if last.cls is ProcProxy and captured == "hiddenobject":
-            # a ProcProxy run using ![] should not be captured
-            return
-    else:
-        cmds_cache = XSH.commands_cache
-        thable = (
-            env.get("THREAD_SUBPROCS")
-            and (captured != "hiddenobject" or env.get("XONSH_CAPTURE_ALWAYS"))
-            and cmds_cache.predict_threadable(last.args)
-            and cmds_cache.predict_threadable(last.cmd)
-        )
-        if captured and thable:
+
+    _last_spec_update_threading(last)
+    _last_spec_update_captured(last)
+
+
+def _last_spec_update_threading(last: SubprocSpec):
+    if callable(last.alias):
+        return
+
+    captured, env, cmds_cache = last.captured, XSH.env, XSH.commands_cache
+    threadable = (
+        captured
+        and env.get("THREAD_SUBPROCS")
+        and (captured != "hiddenobject" or env.get("XONSH_CAPTURE_ALWAYS"))
+        and cmds_cache.predict_threadable(last.args)
+        and cmds_cache.predict_threadable(last.cmd)
+    )
+    if last.force_threadable is not None:
+        threadable = last.force_threadable
+
+    if threadable:
+        if last.captured:
             last.cls = PopenThread
-        elif not thable:
-            # foreground processes should use Popen
-            last.threadable = False
-            if captured == "object" or captured == "hiddenobject":
-                # CommandPipeline objects should not pipe stdout, stderr
-                return
+    else:
+        last.threadable = False
+
+
+def _last_spec_update_captured(last: SubprocSpec):
+    captured = (
+        (captured := last.captured)
+        and not (captured in ["object", "hiddenobject"] and not last.threadable)
+        # a ProcProxy run using ![] should not be captured
+        and not (
+            callable(last.alias)
+            and last.cls is ProcProxy
+            and captured == "hiddenobject"
+        )
+    )
+    if captured:
+        _make_last_spec_captured(last)
+
+
+def _make_last_spec_captured(last: SubprocSpec):
+    captured = last.captured
+    callable_alias = callable(last.alias)
     # cannot used PTY pipes for aliases, for some dark reason,
     # and must use normal pipes instead.
     use_tty = xp.ON_POSIX and not callable_alias
@@ -817,6 +946,64 @@ def _update_last_spec(last):
         last.captured_stderr = last.captured_stdout
 
 
+def _update_proc_alias_threadable(proc):
+    threadable = XSH.env.get("THREAD_SUBPROCS") and getattr(
+        proc.alias, "__xonsh_threadable__", True
+    )
+    if proc.force_threadable is not None:
+        threadable = proc.force_threadable
+    proc.threadable = threadable
+    proc.cls = ProcProxyThread if proc.threadable else ProcProxy
+
+
+def _update_proc_alias_captured(proc):
+    proc.captured = getattr(proc.alias, "__xonsh_capturable__", proc.captured)
+
+
+def _trace_specs(trace_mode, specs, cmds, captured):
+    """Show information about specs."""
+    tracer = XSH.env.get("XONSH_TRACE_SUBPROC_FUNC", None)
+    if callable(tracer):
+        tracer(cmds, captured=captured)
+    else:
+        r = {"cmds": cmds, "captured": captured}
+        print(f"Trace run_subproc({repr(r)})", file=sys.stderr)
+        if trace_mode >= 2:
+            for i, s in enumerate(specs):
+                pcls = s.cls.__module__ + "." + s.cls.__name__
+                pcmd = (
+                    [s.args[0].__name__] + s.args[1:] if callable(s.args[0]) else s.args
+                )
+                p = {
+                    "cmd": pcmd,
+                    "cls": pcls,
+                }
+                p |= {
+                    a: getattr(s, a, None)
+                    for a in [
+                        "alias_name",
+                        "alias",
+                        "binary_loc",
+                        "threadable",
+                        "background",
+                    ]
+                }
+                if trace_mode == 3:
+                    p |= {
+                        a: getattr(s, a, None)
+                        for a in [
+                            "stdin",
+                            "stdout",
+                            "stderr",
+                            "captured",
+                            "captured_stdout",
+                            "captured_stderr",
+                        ]
+                    }
+                p = {k: v for k, v in p.items() if v is not None}
+                print(f"{i}: {repr(p)}", file=sys.stderr)
+
+
 def cmds_to_specs(cmds, captured=False, envs=None):
     """Converts a list of cmds to a list of SubprocSpec objects that are
     ready to be executed.
@@ -849,21 +1036,41 @@ def cmds_to_specs(cmds, captured=False, envs=None):
 
     # Apply boundary conditions
     if not XSH.env.get("XONSH_CAPTURE_ALWAYS"):
-        # Make sure sub-specs are always captured.
-        # I.e. ![some_alias | grep x] $(some_alias)
-        specs_to_capture = specs if captured in STDOUT_CAPTURE_KINDS else specs[:-1]
-        for spec in specs_to_capture:
-            if spec.env is None:
-                spec.env = {"XONSH_CAPTURE_ALWAYS": True}
-            else:
-                spec.env.setdefault("XONSH_CAPTURE_ALWAYS", True)
+        # Make sure sub-specs are always captured in case:
+        # `![some_alias | grep x]`, `$(some_alias)`, `some_alias > file`.
+        last = spec
+        is_redirected_stdout = bool(last.stdout)
+        specs_to_capture = (
+            specs
+            if captured in STDOUT_CAPTURE_KINDS or is_redirected_stdout
+            else specs[:-1]
+        )
+        _set_specs_capture_always(specs_to_capture)
 
     _update_last_spec(specs[-1])
     return specs
 
 
-def _should_set_title():
-    return XSH.env.get("XONSH_INTERACTIVE") and XSH.shell is not None
+def _set_specs_capture_always(specs_to_capture):
+    """Set XONSH_CAPTURE_ALWAYS for all specs."""
+    for spec in specs_to_capture:
+        if spec.env is None:
+            spec.env = {"XONSH_CAPTURE_ALWAYS": True}
+        else:
+            spec.env.setdefault("XONSH_CAPTURE_ALWAYS", True)
+
+
+def _shell_set_title(cmds):
+    if XSH.env.get("XONSH_INTERACTIVE") and XSH.shell is not None:
+        # context manager updates the command information that gets
+        # accessed by CurrentJobField when setting the terminal's title
+        with XSH.env["PROMPT_FIELDS"]["current_job"].update_current_cmds(cmds):
+            # remove current_job from prompt level cache
+            XSH.env["PROMPT_FIELDS"].reset_key("current_job")
+            # The terminal's title needs to be set before starting the
+            # subprocess to avoid accidentally answering interactive questions
+            # from commands such as `rm -i` (see #1436)
+            XSH.shell.settitle()
 
 
 def run_subproc(cmds, captured=False, envs=None):
@@ -879,67 +1086,69 @@ def run_subproc(cmds, captured=False, envs=None):
 
     Lastly, the captured argument affects only the last real command.
     """
-    if XSH.env.get("XONSH_TRACE_SUBPROC", False):
-        tracer = XSH.env.get("XONSH_TRACE_SUBPROC_FUNC")
-        if callable(tracer):
-            tracer(cmds, captured=captured)
-        else:
-            print(f"TRACE SUBPROC: {cmds}, captured={captured}", file=sys.stderr)
 
     specs = cmds_to_specs(cmds, captured=captured, envs=envs)
-    if _should_set_title():
-        # context manager updates the command information that gets
-        # accessed by CurrentJobField when setting the terminal's title
-        with XSH.env["PROMPT_FIELDS"]["current_job"].update_current_cmds(cmds):
-            # remove current_job from prompt level cache
-            XSH.env["PROMPT_FIELDS"].reset_key("current_job")
-            # The terminal's title needs to be set before starting the
-            # subprocess to avoid accidentally answering interactive questions
-            # from commands such as `rm -i` (see #1436)
-            XSH.shell.settitle()
-            # run the subprocess
-            return _run_specs(specs, cmds)
-    else:
-        return _run_specs(specs, cmds)
+
+    if trace_mode := XSH.env.get("XONSH_TRACE_SUBPROC", False):
+        _trace_specs(trace_mode, specs, cmds, captured)
+
+    cmds = [
+        _flatten_cmd_redirects(cmd) if isinstance(cmd, list) else cmd for cmd in cmds
+    ]
+    _shell_set_title(cmds)
+    return _run_specs(specs, cmds)
 
 
-def _run_specs(specs, cmds):
+def _run_command_pipeline(specs, cmds):
     captured = specs[-1].captured
     if captured == "hiddenobject":
-        command = HiddenCommandPipeline(specs)
+        cp = HiddenCommandPipeline(specs)
     else:
-        command = CommandPipeline(specs)
-    proc = command.proc
-    background = command.spec.background
+        cp = CommandPipeline(specs)
+    proc = cp.proc
+    background = cp.spec.background
     if not all(x.is_proxy for x in specs):
         xj.add_job(
             {
                 "cmds": cmds,
-                "pids": [i.pid for i in command.procs],
+                "pids": [i.pid for i in cp.procs],
+                "status": "suspended" if cp.suspended else "running",
                 "obj": proc,
                 "bg": background,
-                "pipeline": command,
-                "pgrp": command.term_pgid,
+                "pipeline": cp,
+                "pgrp": cp.term_pgid,
             }
         )
-    # For some reason, some programs are in a stopped state when the flow
-    # reaches this point, hence a SIGCONT should be sent to `proc` to make
-    # sure that the shell doesn't hang.
-    # See issue #2999 and the fix in PR #3000
+    return cp
+
+
+def _run_specs(specs, cmds):
+    cp = _run_command_pipeline(specs, cmds)
+    XSH.last, proc, captured, background = (
+        cp,
+        cp.proc,
+        specs[-1].captured,
+        cp.spec.background,
+    )
+
+    """
+    For some reason, some programs are in a stopped state when the flow
+    reaches this point, hence a SIGCONT should be sent to `proc` to make
+    sure that the shell doesn't hang. See issue #2999 and the fix in PR #3000
+    """
     resume_process(proc)
 
-    # now figure out what we should return
     if captured == "object":
-        return command  # object can be returned even if backgrounding
+        return cp
     elif captured == "hiddenobject":
         if not background:
-            command.end()
-        return command
+            cp.end()
+        return cp
     elif background:
         return
     elif captured == "stdout":
-        command.end()
-        return command.output
+        cp.end()
+        return cp.output
     else:
-        command.end()
+        cp.end()
         return
